from operator import itemgetter

import boto3

from infra_buddy.aws.cloudformation import CloudFormationBuddy


def _get_all_export_values(client):
    pass


def load_balancer_name(deploy_ctx):
    # type: (DeployContext) -> str
    cf = CloudFormationBuddy(deploy_ctx)
    # we need an export value from the cluster stack so manually override it
    val = _get_cluster_stack_export_value(cf, deploy_ctx,"ElasticLoadBalancerARN")
    # //#amazon conveniently wants some substring of the ARN instead of the name or other value actually available in the API
    # //# turn arn:aws:elasticloadbalancing:us-west-2:271083817914:listener/app/prod-EcsEl-1WYNMMT2MT9NR/c5f92ddeb151227f/313bb2e23d9dd8d8
    # //# into app/prod-EcsEl-1WYNMMT2MT9NR/c5f92ddeb151227f/313bb2e23d9dd8d8
    return "" if not val else val[val.find('app/'):]


def _get_cluster_stack_export_value(cf, deploy_ctx, param):
    # type: (CloudFormationBuddy,DeployContext) -> str
    cf.stack_name = deploy_ctx.cluster_stack_name
    val = cf.get_export_value(param)
    return val


def _get_max_priority(rules):
    rules = sorted(rules, key=lambda k: k["Priority"],reverse=True)
    for rule in rules:
        priority_ = rule['Priority']
        if priority_ is not "default":
            return int(priority_)


def calculate_rule_priority(deploy_ctx,stack_name):
    # type: (DeployContext,str) -> str
    cf = CloudFormationBuddy(deploy_ctx)
    # we need some data for the passed stack_name so manually override it
    cf.stack_name = stack_name
    if cf.does_stack_exist():
        return cf.get_existing_parameter_value('RulePriority')
    else:
        listenerArn = _get_cluster_stack_export_value(cf,deploy_ctx,"ListenerARN")
        client = boto3.client('elbv2',region_name=deploy_ctx.region)
        rules = client.describe_rules(ListenerArn=listenerArn)['Rules']
        if not rules or len(rules)==0:
            current_max = 30
        else:
            if len(rules)==1 and rules[0]['Priority'] == "default":
                current_max = 30
            else:
                current_max = int(_get_max_priority(rules))
        return str(current_max+1)


    # //if [[ $(does_stack_exist ${STACK_NAME}) == "Yes" ]]; then
    # //    RULE_PRIORITY=$(print_stack_param ${STACK_NAME} "RulePriority")
    # //else
    # //    ListenerARN=`print_export_value "$(generate_cluster_stack_name)-ListenerARN"`
    # //    EXISTING_RULES=`aws elbv2 describe-rules --listener-arn ${ListenerARN} | jq ".Rules"`
    # //    if [[ ${EXISTING_RULES} != "null" ]]; then
    # //        CURRENT_MAX=$(aws elbv2 describe-rules --listener-arn ${ListenerARN} | jq ".Rules[].Priority" | sed 's/[^0-9]*//g' | sort -nr | head -n1)
    # //    else
    # //        CURRENT_MAX=30
    # //    fi
    # //    #increment by 30 to allow plenty of room for explicit rule creation
    # //    CURRENT_NUMBER_OF_RULES=$((CURRENT_MAX+1))
    # //fi



import click

# [ --application <otxb>] [ --role <>] [ --environment <>]
from infra_buddy.context.deploy_ctx import DeployContext
from infra_buddy.utility import print_utility


@click.group()
@click.option("--artifact-directory",
              envvar='ARTIFACT_DIRECTORY',
              type=click.Path(exists=True),
              help='A directory where a service definition (service.json) and supporting files can be found.')
@click.option("--application", envvar='APPLICATION', help='The application name.')
@click.option("--role", envvar='ROLE', help='The role name')
@click.option("--environment", envvar='ENVIRONMENT', help='The environment the deployment should target.')
@click.option("--configuration-defaults", envvar='CONFIG_DEFAULTS', type=click.Path(exists=True),
              help='A json file with a dictionary of the default values')
@click.option("--verbose", is_flag=True, help='Print verbose status messages')
@click.pass_context
def cli(ctx, artifact_directory, application, role, environment, configuration_defaults, verbose):
    # type: (object, str, str, str, str, str, bool) -> None
    """
    CLI for managing the infrastructure for deploying micro-services in AWS.
    """
    print_utility.configure(verbose)
    if artifact_directory:
        if application or role:
            raise click.UsageError("When specifying --artifact-directory do not provide --application or --role")
        ctx.obj = DeployContext.create_deploy_context_artifact(artifact_directory=artifact_directory,
                                                               environment=environment,
                                                               defaults=configuration_defaults)
    else:
        ctx.obj = DeployContext.create_deploy_context(application=application,
                                                      role=role,
                                                      environment=environment,
                                                      defaults=configuration_defaults)


# noinspection PyUnresolvedReferences
from infra_buddy.commands.deploy_cloudformation import command
# noinspection PyUnresolvedReferences
from infra_buddy.commands.deploy_service import command
# noinspection PyUnresolvedReferences
from infra_buddy.commands.validate_template import command
# noinspection PyUnresolvedReferences
from infra_buddy.commands.generate_artifact_manifest import command
# noinspection PyUnresolvedReferences
from infra_buddy.commands.generate_service_definition import command
# noinspection PyUnresolvedReferences
from infra_buddy.commands.bootstrap import command
# noinspection PyUnresolvedReferences
from infra_buddy.commands.introspect import command



import json
import os
from collections import defaultdict

import click
from jsonschema import validate

from infra_buddy.template.template import URLTemplate, GitHubTemplate, NamedLocalTemplate, S3Template
from infra_buddy.utility import print_utility


class TemplateManager(object):
    deploy_templates = {}
    service_modification_templates = defaultdict(dict)
    default_service_modification_templates = {}

    schema = {
           "type": "object",
           "additionalProperties": {
               "type": "object",
               "properties": {
                   "type": {"type": "string"},
                   "owner": {"type": "string"},
                   "compatible": {
                               "type": "array",
                               "items": {
                                   "type": "string"
                               },
                               "minItems": 1,
                               "uniqueItems": True
                           },
                   "repo": {"type": "string"},
                   "tag": {"type": "string"},
                   "location": {"type": "string"},
                   "url": {"type": "string"}
               },
               "required":["type"]
           }

       }


    def __init__(self, user_default_service_templates=None, user_default_service_modification_tempaltes=None):
        # type: (DeployContext) -> None
        super(TemplateManager, self).__init__()
        with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'builtin-templates.json'), 'r') as f:
            built_in = json.load(f)
        service_templates_ = built_in['service-templates']
        validate(service_templates_, self.schema)
        self._load_templates(service_templates_)
        modification_templates_ = built_in['service-modification-templates']
        validate(modification_templates_, self.schema)
        self._load_templates(modification_templates_, service_modification=True)
        if user_default_service_templates:self._load_templates(user_default_service_templates)
        if user_default_service_modification_tempaltes:self._load_templates(user_default_service_modification_tempaltes)

    def get_known_service(self, service_type):
        # type: (str) -> Template
        return self.locate_service(service_type)

    def get_known_service_modification(self,service_type, modification_name):
        # type: (str,str) -> Template
        return self.locate_service_modification(service_type=service_type, mod_type=modification_name)

    def get_service_modifications_for_service(self,service_type):
        ret = {}
        ret.update(self.service_modification_templates.get(service_type,{}))
        ret.update(self.default_service_modification_templates)
        return ret

    def get_known_template(self,template_name):
        template = self.deploy_templates.get(template_name,self.default_service_modification_templates.get(template_name))
        if not template:
            for service, template_map in self.service_modification_templates.iteritems():
                template = template_map.get(template_name,None)
        if template:
            template.download_template()
            return template
        else:
            print_utility.error("Unknown service template - {}".format(template_name), raise_exception=True)


    def get_resource_service(self, artifact_directory):
        # type: (str) -> Template
        try:
            template = NamedLocalTemplate(artifact_directory)
            return template
        except click.UsageError as e:
            return None

    def locate_service(self, service_type):
        # type: (str, bool) -> Template
        template = self.deploy_templates.get(service_type, None)
        if not template:
            print_utility.error("Unknown service template - {}".format(service_type), raise_exception=True)
        template.download_template()
        return template

    def locate_service_modification(self, service_type, mod_type):
        # type: (str, str) -> Template
        template = self.service_modification_templates.get(service_type, {}).get(mod_type,None)
        if not template:
            template = self.default_service_modification_templates.get(mod_type,None)
        if not template:
            print_utility.error("Unknown service modification '{}' for type '{}'".format(mod_type,service_type), raise_exception=True)
        template.download_template()
        return template

    def _load_templates(self, templates, service_modification=False):
        # type: (dict, bool) -> None
        for name, values in templates.iteritems():
            type_ = values['type']
            if type_ == "github":
                template = GitHubTemplate(service_type=name, values=values)
            elif type_ == "s3":
                template = S3Template(service_type=name, values=values)
            elif type_ == "url":
                template = URLTemplate(service_type=name, values=values)
            else:
                print_utility.error("Can not locate resource. Requested unknown template type - {}".format(type_),
                                    raise_exception=True)
                raise Exception("")
            if service_modification:
                compatibility = values.get('compatible', [])
                for service in compatibility:
                    if service == "*":
                        self.default_service_modification_templates[name] = template
                    self.service_modification_templates[service][name] = template
            else:
                self.deploy_templates[name] = template





import os
import tempfile
import urlparse
from zipfile import ZipFile

import boto3
import botocore

from infra_buddy.utility import print_utility


class S3Buddy(object):
    def __init__(self, deploy_ctx, root_path, bucket_name):
        super(S3Buddy, self).__init__()
        self.deploy_ctx = deploy_ctx
        self.s3 = boto3.resource('s3', region_name=self.deploy_ctx.region)
        self.bucket = self.s3.Bucket(bucket_name)
        try:
            self.bucket.create(CreateBucketConfiguration={'LocationConstraint': self.deploy_ctx.region})
        except (botocore.exceptions.ValidationError, botocore.exceptions.ClientError) as err:
            pass
        self.bucket_name = bucket_name
        self.deploy_ctx = deploy_ctx
        self.s3 = boto3.resource('s3', region_name=self.deploy_ctx.region)
        self.key_root_path = root_path
        self.url_base = "https://s3-{region}.amazonaws.com/{bucket_name}".format(region=self.deploy_ctx.region,
                                                                                 bucket_name=self.bucket_name)

    def upload(self, file, key_name=None):
        key_name = self._get_upload_bucket_key_name(file, key_name)
        self.bucket.put_object(Key=key_name, Body=open(file, 'rb'))
        print_utility.info("Uploaded file to S3 - Bucket: {} Key: {}".format(self.bucket_name, key_name))
        return "{}/{}".format(self.url_base, key_name)

    def _get_upload_bucket_key_name(self, file, key_name=None):
        key_name = (key_name if key_name else os.path.basename(file))
        if self.key_root_path and self.key_root_path is not '':
            return "{path}/{key_name}".format(path=self.key_root_path,
                                              key_name=key_name)
        return key_name

    def get_file_as_string(self, filename):
        key_name = self._get_upload_bucket_key_name(file=None, key_name=filename)
        return self.s3.meta.client.get_object(Bucket=self.bucket_name, Key=key_name)['Body'].read()


class CloudFormationDeployS3Buddy(S3Buddy):
    def __init__(self, deploy_ctx, ):
        super(CloudFormationDeployS3Buddy, self).__init__(deploy_ctx=deploy_ctx,
                                                          root_path=deploy_ctx.cf_deploy_resource_path,
                                                          bucket_name=deploy_ctx.cf_bucket_name)


def download_zip_from_s3_url(s3_url, destination):
    # type: (str, dest_directory) -> None
    parsed = urlparse.urlparse(s3_url)
    bucket = parsed.hostname
    key = parsed.path[1:]  # strip leading /
    s3 = boto3.resource('s3')
    with tempfile.NamedTemporaryFile() as temporary_file:
        temp_file_path = temporary_file.name
    s3.Bucket(bucket).download_file(key, temp_file_path)
    with ZipFile(temp_file_path) as zf:
        zf.extractall(destination)



import os
import tempfile
from zipfile import ZipFile

import requests

from infra_buddy.aws import s3
from infra_buddy.utility import print_utility


class Template(object):
    def __init__(self, service_type):
        super(Template, self).__init__()
        self.service_type = service_type
        self.destination_relative = None

    def get_parameter_file_path(self):
        return os.path.join(self._get_template_location(),
                            "{service_type}.parameters.json".format(service_type=self.service_type))

    def get_defaults_file_path(self):
        return os.path.join(self._get_template_location(),
                            "defaults.json".format(service_type=self.service_type))

    def get_template_file_path(self):
        return os.path.join(self._get_template_location(), "{service_type}.template".format(service_type=self.service_type))

    def get_config_dir(self):
        config_path = os.path.join(self._get_template_location(), "config")
        return config_path if os.path.exists(config_path) else None

    def _get_template_location(self):
        return os.path.join(self.destination,self.destination_relative) if self.destination_relative else self.destination

    def _validate_template_dir(self):
        if not os.path.exists(self.get_template_file_path()):
            print_utility.error("Template file could not be located for service - {service_type}".format(
                service_type=self.service_type), raise_exception=True)
        if not os.path.exists(self.get_parameter_file_path()):
            print_utility.error("Parameter file could not be located for service - {service_type}".format(
                service_type=self.service_type), raise_exception=True)

    def _prep_download(self):
        self.destination = tempfile.mkdtemp()

    def _set_download_relative_path(self,path):
        self.destination_relative = path





class URLTemplate(Template):
    def __init__(self, service_type, values):
        super(URLTemplate, self).__init__(service_type)
        self.download_url = values.get('url',None)

    def download_template(self):
        self._prep_download()
        r = requests.get(self.download_url, stream=True)
        temporary_file = tempfile.NamedTemporaryFile()
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                temporary_file.write(chunk)
        temporary_file.seek(0)
        with ZipFile(temporary_file) as zf:
            zf.extractall(self.destination)
        self._validate_template_dir()


class GitHubTemplate(URLTemplate):

    def __init__(self, service_type, values):
        super(GitHubTemplate, self).__init__(service_type=service_type,values=values)
        tag = values.pop('tag', 'master')
        self.download_url = "https://github.com/{owner}/{repo}/archive/{tag}.zip".format(tag=tag, **values)
        self._set_download_relative_path("{repo}-{tag}".format(tag=tag,**values))


class NamedLocalTemplate(Template):
    def __init__(self, directory,service_type="aws-resources"):
        super(NamedLocalTemplate, self).__init__(service_type)
        self.destination = directory
        self._validate_template_dir()


class S3Template(Template):
    def __init__(self, service_type, values):
        super(S3Template, self).__init__(service_type)
        self.s3_location = values['location']

    def download_template(self):
        self._prep_download()
        s3.download_zip_from_s3_url(self.s3_location, self.destination)





class LocalTemplate(Template):

    def __init__(self, template, parameter_file, config_dir=None):
        super(LocalTemplate, self).__init__("")
        self.config_dir = config_dir
        self.parameter_file = parameter_file
        self.template = template

    def get_parameter_file_path(self):
        return self.parameter_file

    def get_defaults_file_path(self):
        return None

    def get_template_file_path(self):
        return self.template

    def get_config_dir(self):
        return self.config_dir




import click

from infra_buddy.commandline import cli
from infra_buddy.context.deploy_ctx import DeployContext


@cli.command(name='deploy-service',short_help="Deploy and/or update a service as defined by a service definition "
                                              "and an optional artifact definition.  Operation is idempotent.")
@click.option("--dry-run", is_flag=True, help="Prints the execution plan and displays the evaluated "
                                               "parameter values for the deployment.")
@click.pass_obj
def deploy_cloudformation(deploy_ctx,dry_run):
    # type: (DeployContext,bool) -> None
    do_command(deploy_ctx,dry_run)


def do_command(deploy_ctx, dry_run):
    # type: (DeployContext,bool) -> None
    plan = deploy_ctx.get_execution_plan()
    for deploy in plan:
        deploy.do_deploy(dry_run)



import click

VERBOSE = False


def configure(verbose):
    global VERBOSE
    VERBOSE = verbose


def info(msg):
    if VERBOSE: click.secho(msg, fg='green')


def warn(msg):
    click.secho(msg, fg="yellow")


def banner_warn(msg, data):
    banner(msg)
    info_banner(data)


def error(err_msg,raise_exception=False):
    click.secho(err_msg,fg="red")
    if raise_exception:
        raise click.UsageError(err_msg)


def banner(data):
    click.secho(data, fg="yellow")

def info_banner(data):
    click.secho(data, fg='green')



import logging
import time


def _compare(expected_result, param, negate):
    if type(expected_result) is list:
        if negate:
            return param in expected_result
        else:
            return param not in expected_result

    if negate:
        return expected_result == param
    else:
        return expected_result != param


def waitfor(function_pointer, expected_result, interval_seconds, max_attempts, negate=False, args=None):
    if args is None:
        args = {}
    attempt = 1
    latest = function_pointer(**args)
    while attempt < max_attempts and _compare(expected_result, latest, negate):
        logging.info("Waiting for another attempt - {attempt}".format(attempt=attempt))
        time.sleep(interval_seconds)
        latest = function_pointer(**args)
        attempt += 1
    if attempt >= max_attempts:
        # we failed
        raise Exception("Wait condition failed: {func}".format(func=function_pointer))
    else:
        return latest



import os
import click
from infra_buddy.aws.cloudformation import CloudFormationBuddy
from infra_buddy.aws.s3 import S3Buddy
from infra_buddy.commandline import cli
from infra_buddy.deploy.cloudformation_deploy import CloudFormationDeploy
from infra_buddy.template.template import LocalTemplate


@cli.command(name='deploy-cloudformation')
@click.option('--template', envvar='TEMPLATE', type=click.Path(exists=True),
              help='The cloudformation template to deploy')
@click.option('--parameter-file', envvar='PARAMETER_FILE', type=click.Path(exists=True),
              help='The json parameter file that corresponds with the cloudformation template to deploy')
@click.option('--config-templates', envvar='CONFIG_TEMPLATES', type=click.Path(exists=True),
              help='A directory containing templates to be evaluated and staged for deployment.')
@click.option("--dry-run", is_flag=True, help="Prints the execution plan and displays the evaluated "
                                               "parameter values for the deployment.")
@click.pass_obj
def deploy_cloudformation(deploy_ctx, template, parameter_file, config_templates,dry_run):
    deploy_plan = CloudFormationDeploy(deploy_ctx.stack_name, LocalTemplate(template, parameter_file, config_templates), deploy_ctx)
    deploy_plan.do_deploy(dry_run)
